/*
 * ConsumerSupport.java
 *
 * 作成日: 2008/09/20
 * Copyright(C)2008 Kyushu Institute of Technology., All right reserved. 
 */
package jp.ac.kyutech.SRP.Scripting;

import java.util.Collections;
import java.util.List;

import org.apache.commons.logging.Log;

import jp.go.aist.rtm.RTC.RTObject_impl;
import jp.go.aist.rtm.RTC.port.CorbaConsumer;
import RTC.ReturnCode_t;

/**
 * ConsumerのRTコンポーネントをスクリプトから呼び出すのをサポートします。
 *
 * @author 
 */
public class ConsumerSupport {

    private static Log log = org.apache.commons.logging.LogFactory.getLog(ConsumerSupport.class);
    
    /**
     * 
     */
    protected RTObject_impl component;
    
    /**
     * 
     */
    protected CorbaConsumer<?> serviceBase;
    
    /**
     * アクションを保持するリストです。
     */
    protected List<Action> actionList;
    
    /**
     * デフォルトコンストラクタ。
     */
    public ConsumerSupport() {
        this(null, null);
    }
    
    /**
     * @param componentBase
     * @param serviceBase
     */
    public ConsumerSupport(RTObject_impl componentBase, CorbaConsumer<?> serviceBase) {
        setComponent(componentBase);
        setServiceBase(serviceBase);
        actionList = Collections.synchronizedList(new java.util.ArrayList<Action>());
    }
    
    /**
     * @return component を戻します。
     */
    public RTObject_impl getComponent() {
        return component;
    }

    /**
     * @param component 設定する component
     */
    public void setComponent(RTObject_impl component) {
        this.component = component;
        if (log.isTraceEnabled()) {
            log.trace(String.format("ConsumerSupport: component=%s", 
                    (this.component != null ? this.component.getInstanceName() : "")));
        }
    }

    /**
     * @return actionList を戻します。
     */
    public List<Action> getActionList() {
        return actionList;
    }

    /**
     * @param actionList 設定する actionList
     */
    public void setActionList(List<Action> actionList) {
        this.actionList = actionList;
    }
    
    /**
     * @return serviceBase を戻します。
     */
    public CorbaConsumer<?> getServiceBase() {
        return serviceBase;
    }

    /**
     * @param serviceBase 設定する serviceBase
     */
    public void setServiceBase(CorbaConsumer<?> serviceBase) {
        this.serviceBase = serviceBase;
        if (log.isTraceEnabled()) {
            log.trace(String.format("ConsumerSupport: serviceBase=%s", 
                    (this.serviceBase != null ? this.serviceBase : "")));
        }
    }

    /**
     * スタブを取得します。
     * @return
     */
    public Object getStub() {
        return this.serviceBase._ptr();
    }
    
    /**
     * アクションをキューに追加します。
     * 
     * @param action
     */
    public void append(Action action) {
        this.actionList.add(action);
        if (log.isDebugEnabled()) {
            log.debug(String.format("Append action: type=%s,instance=%s",
                    (this.component != null ? this.component.getTypeName() : "null"),
                    (this.component != null ? this.component.getInstanceName() : "null")));
        }
    }
    
    /**
     * キューに格納されているアクションをクリアします。
     */
    public void clear() {
        this.actionList.clear();
        if (log.isDebugEnabled()) {
            log.debug(String.format("Clear aciton list: type=%s,instance=%s",
                    (this.component != null ? this.component.getTypeName() : "null"),
                    (this.component != null ? this.component.getInstanceName() : "null")));
        }
    }
    
    /**
     * アクションをキューから取り出し、実行します。
     * {@link jp.go.aist.rtm.RTC.RTObject_impl#onExecute(int)}から呼び出されます。
     * 
     * @param ec_id
     * @return
     */
    public ReturnCode_t onExcecute(int ec_id) {
        RTC.ReturnCode_t ret = RTC.ReturnCode_t.RTC_OK; 
        synchronized (this.actionList) {
            if (this.actionList.size() > 0) {
                Action action = this.actionList.remove(0);
                try {
                    ret = action.execute(component, ec_id);
                    if (log.isDebugEnabled()) {
                        log.debug(String.format("Execute action: instance=%s,type=%s,returnCode=%s", this.component.getInstanceName(), this.component.getTypeName(), (ret != null ? ret.value() : "null")));
                    }
                }
                finally {
                    Object lock = action.getLock();
                    if (lock != null) {
                        synchronized (lock) {
                            lock.notify();
                        }
                    }
                }
            }
            else {
                if (log.isTraceEnabled()) {
                    log.trace(String.format("No actions: instance=%s,type=%s,", this.component.getInstanceName(), this.component.getTypeName()));
                }
            }
        }
        if (ret == null) {
            ret = RTC.ReturnCode_t.RTC_OK; 
            log.warn(String.format("ReturnCode is null: instance=%s,type=%s", this.component.getInstanceName(), this.component.getTypeName()));
        }
        return ret;
    }

}
