/*
 * InPortSupport.java
 *
 * 作成日: 2009/10/02
 * Copyright(C)2008 Kyushu Institute of Technology., All right reserved. 
 */
package jp.ac.kyutech.SRP.Scripting;

import javax.swing.event.EventListenerList;

import jp.go.aist.rtm.RTC.RTObject_impl;
import jp.go.aist.rtm.RTC.port.ConnectorDataListenerT;
import jp.go.aist.rtm.RTC.port.InPort;
import jp.go.aist.rtm.RTC.port.ConnectorBase.ConnectorInfo;
import jp.go.aist.rtm.RTC.util.DataRef;

import org.apache.commons.logging.Log;

/**
 * RTコンポーネントにおいて、入力ポートに対するスクリプトから操作をサポートします。<br />
 * 入力ポートにごとにこのオブジェクトが生成されます。
 *
 * @author 
 */
public class InPortSupport {

    private static Log log = org.apache.commons.logging.LogFactory.getLog(InPortSupport.class);

    /**
     * RTコンポーネントを保持します。
     */
    protected RTObject_impl component;
    
    /**
     * 入力ポートを保持します。
     */
    protected InPort<?> inPort;
    
    /**
     * 入力ポートのデータ参照を保持します。
     */
    protected DataRef<?> dataRef;
    
    /**
     * イベントリスナーを保持します。
     */
    protected EventListenerList listenerList;
    
    /**
     * デフォルトコンストラクタ。
     */
    public InPortSupport() {
        
    }
    
    /**
     * RTコンポーネントと入力ポートとデータ参照を指定したコンストラクタです。
     * @param component
     * @param inPort
     * @param dataRef
     */
    public InPortSupport(RTObject_impl component, InPort<?> inPort, DataRef<?> dataRef) {
        this.component = component;
        this.inPort = inPort;
        this.dataRef = dataRef;
        if (this.component == null) {
            throw new IllegalArgumentException("component is null");
        }
        if (this.inPort == null) {
            throw new IllegalArgumentException("inPot is null");
        }
        if (this.dataRef == null) {
            throw new IllegalArgumentException("dataRef is null");
        }
        this.listenerList = new EventListenerList();
    }

    /**
     * @return component を戻します。
     */
    public RTObject_impl getComponent() {
        return component;
    }

    /**
     * @param component 設定する component
     */
    public void setComponent(RTObject_impl component) {
        this.component = component;
    }

    /**
     * @return inPort を戻します。
     */
    public InPort<?> getInPort() {
        return inPort;
    }

    /**
     * @param inPort 設定する inPort
     */
    public void setInPort(InPort<?> inPort) {
        this.inPort = inPort;
    }

    /**
     * @return dataRef を戻します。
     */
    public DataRef<?> getDataRef() {
        return dataRef;
    }

    /**
     * @param dataRef 設定する dataRef
     */
    public void setDataRef(DataRef<?> dataRef) {
        this.dataRef = dataRef;
    }

    /**
     * @return listenerList を戻します。
     */
    public EventListenerList getListenerList() {
        return listenerList;
    }

    /**
     * @param listenerList 設定する listenerList
     */
    public void setListenerList(EventListenerList listenerList) {
        this.listenerList = listenerList;
    }
    
    /**
     * RTコンポーネント用のイベントリスナを追加します。
     * 
     * @param inPortListener 追加するリスナ
     */
    public void addInPortListener(InPortListener inPortListener) {
        this.listenerList.add(InPortListener.class, inPortListener);
    }
    
    /**
     * RTコンポーネント用のイベントリスナを削除します。
     * 
     * @param inPortListener
     */
    public void removeInPortListener(InPortListener inPortListener) {
        this.listenerList.remove(InPortListener.class, inPortListener);
    }
    
    /**
     * RTコンポーネント用のイベントリスナを全て削除します。
     */
    public void clearInPortListener(InPortListener inPortListener) {
        Object[] listenerList = this.listenerList.getListenerList();
        if (listenerList != null) {
            for (Object listener: listenerList) {
                if (listener instanceof InPortListener) {
                    this.listenerList.remove(InPortListener.class, (InPortListener )listener);
                }
            }
        }
    }
    
    /**
     * インポートイベントを通知します。
     * @param e
     */
    public void fireDataReceived(InPortEvent e) {
        if (e == null)
            throw new IllegalArgumentException("event is null");
        
        for (InPortListener listener : listenerList.getListeners(InPortListener.class)) {
            if (listener != null) {
                listener.dataReceived(e);
            }
        }
    }
    
    /**
     * インポートからデータの受け取りをチェックし、必要であればイベントを通知します。
     */
    public void fireDataReceived() {
        if (inPort.isNew()) {
            try {
                inPort.read();
            } catch (Throwable t) {
                log.error("Fail to read inport", t);
                throw new org.apache.commons.lang.exception.NestableRuntimeException("Fail to read inport.", t);
            }
            InPortEvent event = new InPortEvent(this);
            event.setValue(dataRef.v);
            fireDataReceived(event);
        }
    }
    
    /**
     * データポートコネクターのイベントを受取るリスナーを登録します。
     * @param <E>
     * @param type
     * @param clazz
     * @param connectorDataListener
     */
    @SuppressWarnings("unchecked")
    public <E> ConnectorDataListenerT<E> addConnectorDataListener(int type, ConnectorDataListener<E> connectorDataListener) {
        if (dataRef == null) {
            throw new IllegalStateException("dataRef is null");
        }
        Class<E> clazz = (Class<E> )dataRef.v.getClass();
        ConnectorDataListenerT<E> connectorDataListenerT = createConnectorDataListener(clazz, connectorDataListener);
        getInPort().addConnectorDataListener(type, connectorDataListenerT);
        return connectorDataListenerT;
    }
    
    /**
     * データポートコネクターのイベントを受取るリスナーを削除します。
     * @param <E>
     * @param type
     * @param clazz
     * @param connectorDataListener
     */
    @SuppressWarnings("unchecked")
    public <E> ConnectorDataListenerT<E> removeConnectorDataListener(int type, ConnectorDataListenerT<E> connectorDataListenerT) {
        getInPort().removeConnectorDataListener(type, connectorDataListenerT);
        return connectorDataListenerT;
    }
    
    /**
     * データポートコネクターのイベントを受取るリスナーを作成します。
     * @param <E>
     * @param clazz
     * @param connectorDataListener
     * @return
     */
    public <E> ConnectorDataListenerT<E> createConnectorDataListener(Class<E> clazz, final ConnectorDataListener<E> connectorDataListener) {
        ConnectorDataListenerT<E> connectorDataListenerT = new ConnectorDataListenerT<E>(clazz) {
            @Override
            public void operator(ConnectorInfo info, E data){
                connectorDataListener.operator(info, data);
            }
        };
        return connectorDataListenerT;
    }
    
}
