/*
 * OutPortSupport.java
 *
 * 作成日: 2009/10/02
 * Copyright(C)2008 Kyushu Institute of Technology., All right reserved. 
 */
package jp.ac.kyutech.SRP.Scripting;

import java.util.Collections;
import java.util.List;

import jp.go.aist.rtm.RTC.RTObject_impl;
import jp.go.aist.rtm.RTC.port.ConnectorDataListenerT;
import jp.go.aist.rtm.RTC.port.OutPort;
import jp.go.aist.rtm.RTC.port.ConnectorBase.ConnectorInfo;
import jp.go.aist.rtm.RTC.util.DataRef;

import org.apache.commons.logging.Log;

import RTC.ReturnCode_t;

/**
 * RTコンポーネントにおいて、出力ポートに対するスクリプトから操作をサポートします。<br />
 * 出力ポートにごとにこのオブジェクトが生成されます。
 *
 * @author 
 */
public class OutPortSupport {

    private static Log log = org.apache.commons.logging.LogFactory.getLog(OutPortSupport.class);

    /**
     * RTコンポーネントを保持します。
     */
    protected RTObject_impl component;
    
    /**
     * 入力ポートを保持します。
     */
    protected OutPort<?> outPort;
    
    /**
     * 入力ポートのデータ参照を保持します。
     */
    protected DataRef<?> dataRef;
    
    /**
     * アクションを保持するリストです。
     */
    protected List<Action> actionList;
    
    /**
     * デフォルトコンストラクタ。
     */
    public OutPortSupport() {
        
    }
    
    /**
     * RTコンポーネントと入力ポートとデータ参照を指定したコンストラクタです。
     * @param component
     * @param outPort
     * @param dataRef
     */
    public OutPortSupport(RTObject_impl component, OutPort<?> outPort, DataRef<?> dataRef) {
        this.component = component;
        this.outPort = outPort;
        this.dataRef = dataRef;
        if (this.component == null) {
            throw new IllegalArgumentException("component is null");
        }
        if (this.outPort == null) {
            throw new IllegalArgumentException("outPot is null");
        }
        if (this.dataRef == null) {
            throw new IllegalArgumentException("dataRef is null");
        }
        this.actionList = Collections.synchronizedList(new java.util.ArrayList<Action>());
    }

    /**
     * @return component を戻します。
     */
    public RTObject_impl getComponent() {
        return component;
    }

    /**
     * @param component 設定する component
     */
    public void setComponent(RTObject_impl component) {
        this.component = component;
    }

    /**
     * @return outPort を戻します。
     */
    public OutPort<?> getOutPort() {
        return outPort;
    }

    /**
     * @param outPort 設定する outPort
     */
    public void setOutPort(OutPort<?> outPort) {
        this.outPort = outPort;
    }

    /**
     * @return dataRef を戻します。
     */
    public DataRef<?> getDataRef() {
        return dataRef;
    }

    /**
     * @param dataRef 設定する dataRef
     */
    public void setDataRef(DataRef<?> dataRef) {
        this.dataRef = dataRef;
    }

    /**
     * @return actionList を戻します。
     */
    public List<Action> getActionList() {
        return actionList;
    }

    /**
     * @param actionList 設定する actionList
     */
    public void setActionList(List<Action> actionList) {
        this.actionList = actionList;
    }

    /**
     * アクションをキューに追加します。
     * 
     * @param action
     */
    public void append(Action action) {
        this.actionList.add(action);
        if (log.isDebugEnabled()) {
            log.debug(String.format("Append action: type=%s,instance=%s",
                    (this.component != null ? this.component.getTypeName() : "null"),
                    (this.component != null ? this.component.getInstanceName() : "null")));
        }
    }
    
    /**
     * キューに格納されているアクションをクリアします。
     */
    public void clear() {
        this.actionList.clear();
        if (log.isDebugEnabled()) {
            log.debug(String.format("Clear aciton list: type=%s,instance=%s",
                    (this.component != null ? this.component.getTypeName() : "null"),
                    (this.component != null ? this.component.getInstanceName() : "null")));
        }
    }
    
    /**
     * アクションをキューから取り出し、実行します。
     * {@link jp.go.aist.rtm.RTC.RTObject_impl#onExecute(int)}から呼び出されます。
     * 
     * @param ec_id
     * @return
     */
    public ReturnCode_t onExcecute(int ec_id) {
        RTC.ReturnCode_t ret = RTC.ReturnCode_t.RTC_OK; 
        synchronized (this.actionList) {
            if (this.actionList.size() > 0) {
                Action action = this.actionList.remove(0);
                try {
                    ret = action.execute(component, ec_id);
                    if (log.isDebugEnabled()) {
                        log.debug(String.format("Execute action: instance=%s,type=%s,returnCode=%s", this.component.getInstanceName(), this.component.getTypeName(), (ret != null ? ret.value() : "null")));
                    }
                }
                finally {
                    Object lock = action.getLock();
                    if (lock != null) {
                        synchronized (lock) {
                            lock.notify();
                        }
                    }
                }
            }
            else {
                if (log.isTraceEnabled()) {
                    log.trace(String.format("No actions: instance=%s,type=%s,", this.component.getInstanceName(), this.component.getTypeName()));
                }
            }
        }
        if (ret == null) {
            ret = RTC.ReturnCode_t.RTC_OK; 
            log.warn(String.format("ReturnCode is null: instance=%s,type=%s", this.component.getInstanceName(), this.component.getTypeName()));
        }
        return ret;
    }

    /**
     * データポートコネクターのイベントを受取るリスナーを登録します。
     * @param <E>
     * @param type
     * @param clazz
     * @param connectorDataListener
     */
    @SuppressWarnings("unchecked")
    public <E> ConnectorDataListenerT<E> addConnectorDataListener(int type, ConnectorDataListener<E> connectorDataListener) {
        if (dataRef == null) {
            throw new IllegalStateException("dataRef is null");
        }
        Class<E> clazz = (Class<E> )dataRef.v.getClass();
        ConnectorDataListenerT<E> connectorDataListenerT = createConnectorDataListener(clazz, connectorDataListener);
        getOutPort().addConnectorDataListener(type, connectorDataListenerT);
        return connectorDataListenerT;
    }
    
    /**
     * データポートコネクターのイベントを受取るリスナーを削除します。
     * @param <E>
     * @param type
     * @param clazz
     * @param connectorDataListener
     */
    @SuppressWarnings("unchecked")
    public <E> ConnectorDataListenerT<E> removeConnectorDataListener(int type, ConnectorDataListenerT<E> connectorDataListenerT) {
        getOutPort().removeConnectorDataListener(type, connectorDataListenerT);
        return connectorDataListenerT;
    }
    
    /**
     * データポートコネクターのイベントを受取るリスナーを作成します。
     * @param <E>
     * @param clazz
     * @param connectorDataListener
     * @return
     */
    public <E> ConnectorDataListenerT<E> createConnectorDataListener(Class<E> clazz, final ConnectorDataListener<E> connectorDataListener) {
        ConnectorDataListenerT<E> connectorDataListenerT = new ConnectorDataListenerT<E>(clazz) {
            @Override
            public void operator(ConnectorInfo info, E data){
                connectorDataListener.operator(info, data);
            }
        };
        return connectorDataListenerT;
    }
    
}
